@@ -129,10 +129,10 @@ func main() {
 		laddr.Port = 0
 		transport, ok := http.DefaultTransport.(*http.Transport)
 		if ok {
-			transport.DialContext = (&net.Dialer{
+			transport.Dial = (&net.Dialer{
 				Timeout:   30 * time.Second,
 				LocalAddr: laddr,
-			}).DialContext
+			}).Dial
 		}
 	}
 
